﻿#region File Description
//-----------------------------------------------------------------------------
// InstancedModelProcessor.cs
//
// Microsoft XNA Community Game Platform
// Copyright (C) Microsoft Corporation. All rights reserved.
// このソースコードはクリエータークラブオンラインのMeshInstancingの
// プロセッサーコードのコメントを翻訳、変更したもの
// http://creators.xna.com/en-US/sample/meshinstancing
//-----------------------------------------------------------------------------
#endregion

#region Using ステートメント

using System;
using System.Collections.Generic;
using System.ComponentModel;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline.Processors;

#endregion

namespace DemoPipeline
{
    /// <summary>
    /// インスタンスモデルプロセッサー
    /// NodeContentからInstanceModelContentに変換する
    /// </summary>
    [ContentProcessor( DisplayName = "インスタンスモデルプロセッサ" )]
    public class InstancedModelProcessor : ContentProcessor<NodeContent,
                                                            InstancedModelContent>
    {
        #region プロパティ

        /// <summary>
        /// ノード変換行列をベイクするか 
        /// </summary>
        [DefaultValue( true )]
        [DisplayName( "ノードトランスフォームのベイク" )]
        [Description( "ノードトランスフォーム情報をモデルの頂点データにベイクするか。" )]
        public bool BakeTransform { get; set; }

        /// <summary>
        /// エフェクトファイルの指定
        /// </summary>
        [DisplayName( "エフェクトファイル" )]
        [Description( "使用するエフェクトファイルを指定する。" )]
        public string EffectFilename { get; set; }

        #endregion

        #region フィールド

        protected NodeContent rootNode;
        protected ContentProcessorContext context;
        protected InstancedModelContent outputModel;

        // マテリアルは複数のジオメトリ間で共有される事があるので
        // 同じマテリアルが再変換されないように、
        // この辞書に変換されたマテリアルを記憶しておく。
        Dictionary<MaterialContent, MaterialContent> processedMaterials =
                            new Dictionary<MaterialContent, MaterialContent>();

        #endregion

        public InstancedModelProcessor() : base()
        {
            // プロパティの規定値の設定
            BakeTransform = true;
        }

        /// <summary>
        /// NodeContentからInstanceModelContentに変換する
        /// </summary>
        public override InstancedModelContent Process( NodeContent input,
                                                      ContentProcessorContext context )
        {
            if ( String.IsNullOrEmpty(EffectFilename) )
            {
                EffectFilename = "InstancedModel.fx";
            }

            this.rootNode = input;
            this.context = context;

            outputModel = new InstancedModelContent();

            ProcessNode( input );

            return outputModel;
        }


        /// <summary>
        /// 再帰的にNodeContentツリーを処理する
        /// </summary>
        void ProcessNode( NodeContent node )
        {
            // ノードトランスフォーム情報をベイクする
            if ( BakeTransform )
            {
                MeshHelper.TransformScene( node, node.Transform );
                node.Transform = Matrix.Identity;
            }

            // このノードはメッシュか？
            MeshContent mesh = node as MeshContent;

            if ( mesh != null )
                ProcessMesh( mesh );

            // 子ノードを処理する
            foreach ( NodeContent child in node.Children )
            {
                ProcessNode( child );
            }
        }

        void ProcessMesh( MeshContent mesh )
        {
            // 頂点とインデックスを描画時のGPUキャッシュヒット率を高める為に並び替える
            MeshHelper.OptimizeForCache( mesh );

            // メッシュ内の全てのジオメトリを処理する
            foreach ( GeometryContent geometry in mesh.Geometry )
            {
                ProcessGeometry( geometry );
            }
        }

        /// <summary>
        /// Geometryからインスタンス用フォーマットに変換する
        /// </summary>
        void ProcessGeometry( GeometryContent geometry )
        {
            int indexCount = geometry.Indices.Count;
            int vertexCount = geometry.Vertices.VertexCount;

            // 指定されたジオメトリの頂点数がインスタンス用に適したものか調べる
            if ( vertexCount > ushort.MaxValue )
            {
                throw new InvalidContentException(
                    string.Format( "ジオメトリの使用頂点数 {0}: " +
                            "インスタンス用には多すぎる頂点数です。", vertexCount ) );
            }

            if ( vertexCount > ushort.MaxValue / 8 )
            {
                context.Logger.LogWarning( null, rootNode.Identity,
                                "ジオメトリの使用頂点数 {0}: " +
                                "このジオメトリはバッチ内で{1}回しか描画できません。" +
                                "より少ない頂点数の方が効果的です。",
                                vertexCount, ushort.MaxValue / vertexCount );
            }

            // 実行時にインスタンスデータを格納するテクスチャ座標が
            // ジオメトリの頂点チャンネルに使われていないかチェックする
            // 注:このチェックはシェーダーインスタンス時には必要だが
            // HWインスタンス時には必要ない。
            VertexChannelCollection vertexChannels = geometry.Vertices.Channels;

            for ( int i = 1; i <= 4; i++ )
            {
                if ( vertexChannels.Contains( VertexChannelNames.TextureCoordinate(i) ) )
                {
                    throw new InvalidContentException(
                        string.Format( "このモデルのテクスチャ座標チャンネル{0}が" +
                                      "既に使われています。インスタンスプログラムは" +
                                      "このチャンネルを使用します。", i ) );
                }
            }

            // 頂点チャンネル(VertexContent)から頂点バッファへの変換
            VertexBufferContent vertexBufferContent;
            VertexElement[] vertexElements;

            geometry.Vertices.CreateVertexBuffer( out vertexBufferContent,
                                                 out vertexElements,
                                                 context.TargetPlatform );

            int vertexStride =
                        VertexDeclaration.GetVertexStrideSize( vertexElements, 0 );

            // マテリアルの変換
            MaterialContent material = ProcessMaterial( geometry.Material );

            // 出力モデルに変換したジオメトリを追加する
            outputModel.AddModelPart( indexCount, vertexCount, vertexStride,
                                     vertexElements, vertexBufferContent,
                                     geometry.Indices, material );
        }

        /// <summary>
        /// 入力マテリアルをインスタンスモデル用のエフェクトマテリアルに変換する
        /// </summary>
        MaterialContent ProcessMaterial( MaterialContent material )
        {
            // このマテリアルは既に変換したか？
            if ( !processedMaterials.ContainsKey( material ) )
            {
                // 入力マテリアルからエフェクトマテリアルへの変換
                EffectMaterialContent instancedMaterial = AddEffectMaterial( material );

                // MaterialProcessorを使ってマテリアルをビルドする。
                // ここでは参照されているエフェクトやテクスチャのビルドも含まれる
                processedMaterials[material] =
                    context.Convert<MaterialContent,
                                    MaterialContent>( instancedMaterial,
                                                     "MaterialProcessor" );
            }

            return processedMaterials[material];
        }

        /// <summary>
        /// 入力マテリアルからエフェクトマテリアルへの変換
        /// </summary>
        protected virtual EffectMaterialContent AddEffectMaterial(
                                                            MaterialContent material )
        {
            EffectMaterialContent instancedMaterial = new EffectMaterialContent();

            // エフェクトファイルを設定する
            instancedMaterial.Effect = new ExternalReference<EffectContent>(
                                            EffectFilename, rootNode.Identity );

            // テクスチャの設定を入力マテリアルからコピーする
            if ( !material.Textures.ContainsKey( "Texture" ) )
            {
                throw new InvalidContentException(
                    "マテリアルにテクスチャが設定されていません。" +
                    "インスタンスモデルにはテクスチャなしの" +
                    "マテリアルはサポートされていません。" );
            }

            instancedMaterial.Textures.Add( "BasicTexture",
                                                material.Textures["Texture"] );

            // その他のパラメーター設定をコピーする
            string[] paramNames =
            {
                "Alpha",
                "DiffuseColor",
                "EmissiveColor",
                "SpecularColor",
                "SpecularPower"
            };

            foreach ( string paramName in paramNames )
            {
                if ( material.OpaqueData.ContainsKey( paramName ) )
                {
                    instancedMaterial.OpaqueData.Add(
                                            paramName, material.OpaqueData[paramName] );
                }
            }

            return instancedMaterial;
        }
    }
}
