﻿#region Using ステートメント

using System;
using System.Collections;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;

#endregion

namespace DemoPipeline
{
    /// <summary>
    /// 複数のメッシュを単一のメッシュに変換する。
    /// 今回のデモでは複数のメッシュノードを１つのメッシュとすることで
    /// インスタンスモデルとして描画するようになっている。
    /// 他にも部屋のモデルを3Dモデリングツール上で作る場合には
    /// 複数のノードになっているが、ゲーム中にそれらのノードが
    /// 動かない場合などには１つのモデルとして描画した方が効率的になる。
    /// </summary>
    public class MeshPacker
    {
        // 処理中のメッシュ
        MeshContent mesh = new MeshContent();

        /// <summary>
        /// メッシュを追加する
        /// </summary>
        public void AddMesh( MeshContent addingMesh )
        {
            int basePositionIndex = mesh.Positions.Count;

            // メッシュの座標データを追加する
            foreach ( Vector3 position in addingMesh.Positions )
                mesh.Positions.Add( position );

            foreach ( GeometryContent g in addingMesh.Geometry )
                AddGeometry( g, basePositionIndex );
        }

        /// <summary>
        /// パック作業を終了し、パックされたメッシュを返す
        /// </summary>
        public MeshContent FinishMesh()
        {
            MeshContent result = mesh;
            mesh = null;
            return result;
        }

        /// <summary>
        /// ジオメトリの追加
        /// </summary>
        private void AddGeometry( GeometryContent addingGeometry, int basePositionIndex )
        {
            GeometryContent geometry = null;

            // 既に追加されているジオメトリと同じタイプか？
            foreach ( GeometryContent g in mesh.Geometry )
            {
                if ( IsSameType( g, addingGeometry ) )
                {
                    geometry = g;
                    break;
                }
            }

            // 新しいタイプのジオメトリとして追加する
            bool isNewGeometry = false;
            if ( geometry == null )
            {
                geometry = new GeometryContent();

                geometry.Name = addingGeometry.Name;
                geometry.Material = addingGeometry.Material;

                mesh.Geometry.Add( geometry );

                isNewGeometry = true;
            }

            int baseIndex = geometry.Vertices.VertexCount;

            // ジオメトリデータを追加する
            if ( isNewGeometry )
            {
                // 座標の追加
                foreach ( int positionIndex in addingGeometry.Vertices.PositionIndices )
                {
                    geometry.Vertices.Add( basePositionIndex + positionIndex );
                }

                // 頂点チャンネルの追加
                for ( int i = 0; i < addingGeometry.Vertices.Channels.Count; ++i )
                {
                    VertexChannel sourceChannel = addingGeometry.Vertices.Channels[i];
                    geometry.Vertices.Channels.Add(
                        sourceChannel.Name, sourceChannel.ElementType, sourceChannel );
                }
            }
            else
            {
                // 既存の頂点チャンネルに追加
                // 現在のチャンネルデータを一旦保存して、チャンネルをクリアした後に
                // 新しいチャンネルデータを追加してチャンネルを構成する
                int channelCount = geometry.Vertices.Channels.Count;
                ArrayList[] channelData = new ArrayList[channelCount];
                for ( int i = 0; i < channelCount; ++i )
                    channelData[i] = new ArrayList( geometry.Vertices.Channels[i] );

                geometry.Vertices.Channels.Clear();

                // ジオメトリデータを追加する
                foreach ( int positionIndex in addingGeometry.Vertices.PositionIndices )
                    geometry.Vertices.Add( basePositionIndex + positionIndex );

                for ( int i = 0; i < channelCount; ++i )
                {
                    VertexChannel sourceChannel = addingGeometry.Vertices.Channels[i];
                    channelData[i].AddRange( sourceChannel );

                    geometry.Vertices.Channels.Add(
                        sourceChannel.Name, sourceChannel.ElementType, channelData[i] );
                }
            }

            // インデックスデータを追加
            foreach ( int index in addingGeometry.Indices )
                geometry.Indices.Add( baseIndex + index );
        }


        /// <summary>
        /// 指定された２つのジオメトリは同一か調べるヘルパーメソッド
        /// </summary>
        static bool IsSameType( GeometryContent a, GeometryContent b )
        {
            // 同じマテリアルを使っているか？
            if ( a.Material != b.Material )
                return false;

            // 頂点チャンネルが同じものか？
            if ( a.Vertices.Channels.Count != b.Vertices.Channels.Count )
                return false;

            // 各チャンネルのタイプが同一なものか調べる
            for ( int i = 0; i < a.Vertices.Channels.Count; ++i )
            {
                VertexChannel cha = a.Vertices.Channels[i];
                VertexChannel chb = b.Vertices.Channels[i];

                if ( cha.ElementType != chb.ElementType )
                    return false;

                if ( cha.Name != chb.Name )
                    return false;
            }

            return true;
        }

    }
}
