﻿#region File Description
//-----------------------------------------------------------------------------
// InstancedModelPart.cs
//
// Microsoft XNA Community Game Platform
// Copyright (C) Microsoft Corporation. All rights reserved.
// このソースコードはクリエータークラブオンラインのMeshInstancingの
// コードのコメントを翻訳、変更したもの
// http://creators.xna.com/en-US/sample/meshinstancing
//-----------------------------------------------------------------------------
#endregion

#region Using ステートメント

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;

#endregion

namespace DemoTypes
{
    /// <summary>
    /// InstancedModelの殆どの機能を実装しているInstanceModelPartクラス
    /// 
    /// InstancedModelは単なるInstanceModelPartの集まりで、InstanceModelPartは
    /// それぞれに別々のマテリアルを指定することができ、複数のインスタンス手法を用いて
    /// 実際にインスタンスモデル描画をする。
    /// 
    /// このデモではパーシャルクラスを使用して、それぞれの初期化コードとパラメーターを
    /// HWInstanceePart.cs、ShaderInstancePart.csの別々のファイルに記述している。
    /// </summary>
    public partial class InstancedModelPart
    {
        #region プロパティ

        /// <summary>
        /// １つの描画バッチで描画できる最大インスタンス数の取得
        /// シェーダーインスタンスの場合はシェーダーで使える定数レジスタの数に左右される。
        /// また、シェーダーインスタンス、HWインスタンスのどちらの場合も
        /// 描画対象となるモデルの頂点数によって変換する。
        /// </summary>
        public int MaxInstances { get; private set; }

        /// <summary>
        /// 使用するエフェクトの取得と設定
        /// </summary>
        public Effect Effect { get; private set; }

        #endregion

        #region フィールド

        // モデルデータ
        int indexCount;
        int vertexCount;
        int vertexStride;

        VertexDeclaration vertexDeclaration;
        VertexBuffer vertexBuffer;
        IndexBuffer indexBuffer;

        InstancedModel owner;

        /// <summary>
        /// シェーダーインスタンス、HWインスタンスのどちらの場合も頂点宣言を変更するので
        /// 変更前の頂点宣言要素を保存しておく必要がある。
        /// </summary>
        VertexElement[] originalVertexDeclaration;

        #endregion

        #region 初期化

        /// <summary>
        /// コンストラクタ
        /// XNBフォーマットからインスタンスモデルを読み込む。
        /// </summary>
        internal InstancedModelPart( InstancedModel owner, ContentReader input,
                                        GraphicsDevice graphicsDevice )
        {
            this.owner = owner;

            // モデルデータの読み込み
            indexCount = input.ReadInt32();
            vertexCount = input.ReadInt32();
            vertexStride = input.ReadInt32();

            vertexDeclaration = input.ReadObject<VertexDeclaration>();
            vertexBuffer = input.ReadObject<VertexBuffer>();
            indexBuffer = input.ReadObject<IndexBuffer>();

            input.ReadSharedResource<Effect>( delegate( Effect value )
            {
                Effect = value;
            } );

            // オリジナルの頂点宣言を保存する
            originalVertexDeclaration = vertexDeclaration.GetVertexElements();
        }


        /// <summary>
        /// モデルパートを指定したインスタンス手法用に初期化する。
        /// このメソッドはスタートアップ時に呼ばれ、後はインスタンス手法を変更した時に
        /// 何度でも自由に呼ぶことができる。
        /// </summary>
        internal void Initialize( InstancingTechnique instancingTechnique,
                                    VertexElement[] instanceVertexElements )
        {
            switch ( instancingTechnique )
            {
            case InstancingTechnique.ShaderInstancing:
                InitializeShaderInstancing( false );
                Effect.CurrentTechnique = Effect.Techniques["ShaderInstancing"];
                break;
            /*
            case InstancingTechnique.HardwareInstancing:
                InitializeHardwareInstancing( instanceVertexElements );
                Effect.CurrentTechnique = Effect.Techniques["HardwareInstancing"];
                break;
            */
            }
        }


        #endregion

        #region 描画

        /// <summary>
        /// ヘルパーメソッド
        /// GraphicsDeviceとEffectを設定し、インスタンスジオメトリを描画する準備をする
        /// </summary>
        public void Begin( Matrix view, Matrix projection, Vector3 eyePosition )
        {
            GraphicsDevice gd = vertexBuffer.GraphicsDevice;

            switch ( owner.InstancingTechnique )
            {
            case InstancingTechnique.ShaderInstancing:

                gd.Vertices[0].SetSource(
                                    shaderInstanceVB, 0, shaderInstanceVertexStride );

                gd.Indices = shaderInstanceIB;
                gd.VertexDeclaration = shaderInstanceVDecl;
                break;

            case InstancingTechnique.HardwareInstancing:

                gd.Vertices[0].SetSource( vertexBuffer, 0, vertexStride );
                gd.Indices = indexBuffer;
                gd.VertexDeclaration = vertexDeclaration;
#if XBOX360
                Effect.Parameters["VertexCount"].SetValue(indexCount);
#endif
                break;
            }

            // カメラ関連の行列を設定する
            Effect.Parameters["View"].SetValue( view );
            Effect.Parameters["Projection"].SetValue( projection );
            Effect.Parameters["EyePosition"].SetValue( eyePosition );

            // エフェクトの開始
            Effect.Begin();
            Effect.CurrentTechnique.Passes[0].Begin();
        }

        /// <summary>
        /// モデルインスタンスの描画
        /// </summary>
        /// <param name="instanceCount">描画するインスタンス数</param>
        public void Draw( int instanceCount )
        {
            GraphicsDevice gd = vertexDeclaration.GraphicsDevice;

            switch ( owner.InstancingTechnique )
            {
            case InstancingTechnique.ShaderInstancing:

                gd.DrawIndexedPrimitives(
                    PrimitiveType.TriangleList, 0, 0,
                    vertexCount * instanceCount, 0,
                    instanceCount * indexCount / 3 );

                break;

            case InstancingTechnique.HardwareInstancing:
                {
#if XBOX360
                    // Xbox 360上ではvfetchを使っているのでインタンス数ではなく
                    // 描画する三角形の総数を渡す
                    int primitiveCount = instanceCount * indexCount / 3;
#else
                    // Windows上で指定するのは１つのインスタンス当たりの三角形数
                    int primitiveCount = indexCount / 3;

                    gd.Vertices[0].SetFrequencyOfIndexData( instanceCount );
                    gd.Vertices[1].SetFrequencyOfInstanceData( 1 );
#endif
                    // １つのバッチで全てのインスタンスを描画する
                    gd.DrawIndexedPrimitives( PrimitiveType.TriangleList,
                                                 0, 0, vertexCount,
                                                 0, primitiveCount );
                }
                break;
            }
        }

        /// <summary>
        /// 描画終了
        /// Effect.Endなどを呼び出す。
        /// </summary>
        public void End()
        {
            GraphicsDevice gd = vertexBuffer.GraphicsDevice;

            // インスタンス用のストリームをリセットする
            // 一部のビデオカードではこのメソッドを呼ばないと正しく描画されないことがある
            gd.Vertices[0].SetSource( null, 0, 0 );
            gd.Vertices[1].SetSource( null, 0, 0 );

            Effect.CurrentTechnique.Passes[0].End();
            Effect.End();
        }

        #endregion
    }
}
