﻿using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using System.Runtime.InteropServices;   // StructLayoutアトリビュートの為

namespace ShaderInstancingTest03
{
    /// <summary>
    /// ゲームオブジェクト構造体
    /// </summary>
    [StructLayout( LayoutKind.Sequential )]
    public struct GameObject
    {
        // アニメ更新間隔の宣言（ミリ秒）
        private const int animeUpdateInterval = 80;

        // モデル回転速度
        private const float modelRotSpeed = 10.0f;


        // CPUで更新し、GPUで読み込まれる変数
        public Vector3  Position;       // 座標
        public float    Scale;          // スケール
        // public Vector3  RotateAxis;     // 回転軸
        // public float    Rotation;       // 回転
        public float Rot;                   // 回転
        public float AnimeTimer;            // アニメタイマー
        public int AnimeFlameNo;            // アニメ表示フレーム番号
        public Vector2 TextureCoordinate;   // テクスチャ座標


        // CPUのみで使用する変数
        public Vector3  Velocity;       // 速度
        public float    RotationSpeed;  // 回転速度
        public float    LifeTime;       // 寿命



        // テクスチャの宣言
        // 画像１枚当たりの横幅（float）の算出に使用している
        private Texture2D myTexture;



        /// <summary>
        /// ゲームオブジェクトの初期化
        /// </summary>
        public void Initialize( Random rnd, Texture2D myTexture )
        {
            this.myTexture = myTexture;

            // 出現位置、移動速度などをランダムに生成する
            float w = Game1.Sandbox.Max.X - Game1.Sandbox.Min.X;
            float h = Game1.Sandbox.Max.Y - Game1.Sandbox.Min.Y;

            Position = new Vector3(
                Game1.Sandbox.Min.X + w * (float)rnd.NextDouble(),
                Game1.Sandbox.Min.Y + h * (float)rnd.NextDouble(),
                (float)rnd.NextDouble() * -50.0f );

            float rad = (float)rnd.NextDouble() * MathHelper.TwoPi;
            Velocity = new Vector3( (float)Math.Cos( rad ),
                                    (float)Math.Sin( rad ), 0 ) * 3.0f;

            //float theta = (float)rnd.NextDouble() * MathHelper.TwoPi;
            //float phi = (float)rnd.NextDouble() * MathHelper.TwoPi;
            //RotateAxis = Vector3.Transform( Vector3.UnitX,
            //                    Matrix.CreateFromYawPitchRoll( theta, 0, phi ) );
            //Rotation = (float)rnd.NextDouble() * MathHelper.TwoPi;
            RotationSpeed = (float)rnd.NextDouble() * MathHelper.TwoPi;

            Scale = 0.5f + (float)rnd.NextDouble();
            LifeTime = 20.0f + 10.0f * (float)rnd.NextDouble();

            Rot = 0;

            // タイマーの初期値をランダムに設定（アニメ更新間隔×横枚数×縦枚数）
            AnimeTimer = (float)rnd.Next(animeUpdateInterval * 25 * 24);
            AnimeFlameNo = 0;

            // テクスチャ座標を初期化
            TextureCoordinate = new Vector2(0, 0);
        }

        /// <summary>
        /// ゲームオブジェクトの更新
        /// </summary>
        /// <param name="dt">更新時間</param>
        /// <returns>寿命が尽きたらfalseを返す</returns>
        public bool Update(GameTime gameTime, float dt)
        {
            // タイマーの更新
            AnimeTimer += (float)gameTime.ElapsedGameTime.TotalMilliseconds;

            // 新表示アニメ番号を算出
            int newAnimeFlameNo = (int)(AnimeTimer / animeUpdateInterval);
            if (newAnimeFlameNo >= 25 * 24)
            {
                newAnimeFlameNo = 0;
                AnimeTimer = 0;
            }

            // 前回の表示アニメ番号と新表示アニメ番号を比較チェック
            if (AnimeFlameNo != newAnimeFlameNo)
            {
                // 表示アニメ番号を更新
                AnimeFlameNo = newAnimeFlameNo;

                // テクスチャ座標を更新する
                this.textureCoordinateUpdate(AnimeFlameNo);
            }

            // 板ポリを回転させる
            Rot = (AnimeTimer / (animeUpdateInterval * 25 * 24)) * (float)MathHelper.TwoPi * modelRotSpeed;



            //// 座標と回転の更新
            //Position.X += Velocity.X * dt;
            //Position.Y += Velocity.Y * dt;
            //Position.Z += Velocity.Z * dt;
            //Rotation = FastMath.NormalizeRotation( Rotation + RotationSpeed * dt );

            //// 移動範囲の端に来たら跳ね返る処理
            //if ((Position.X < Game1.Sandbox.Min.X && Velocity.X < 0) ||
            //        (Position.X > Game1.Sandbox.Max.X && Velocity.X > 0))
            //{
            //    Velocity.X = -Velocity.X;
            //}

            //if ((Position.Y < Game1.Sandbox.Min.Y && Velocity.Y < 0) ||
            //    (Position.Y > Game1.Sandbox.Max.Y && Velocity.Y > 0))
            //{
            //    Velocity.Y = -Velocity.Y;
            //}

            //// 寿命の更新
            //LifeTime -= dt;

            return LifeTime > 0.0f;
        }

        /// <summary>
        /// 頂点要素
        /// </summary>
        static public VertexElement[] VertexElements = {
            // Position (8バイト)
            new VertexElement(0, 0, VertexElementFormat.Vector2,
                                    VertexElementMethod.Default,
                                    VertexElementUsage.Position, 0),
            // Color  (4バイト)
            new VertexElement(0, 8, VertexElementFormat.Color,
                                    VertexElementMethod.Default,
                                    VertexElementUsage.Color, 0),
            // Scale, Rotation, Depth (12バイト)
            new VertexElement(0,12, VertexElementFormat.Vector3,
                                    VertexElementMethod.Default,
                                    VertexElementUsage.TextureCoordinate, 0),
        };

        /// <summary>
        /// この構造体のサイズを返す
        /// </summary>
        public static int SizeInBytes
        {
            get { return Marshal.SizeOf( typeof( GameObject ) ); }
        }


        // テクスチャ座標を更新する
        private void textureCoordinateUpdate(int animeNo)
        {
            // テクスチャ座標を算出
            int iWidth = 80;    // 画像１枚当たりの横幅（int）
            int iHeight = 80;   // 画像１枚当たりの縦幅（int）
            int animeWidthMax = 25;    // アニメ画像の横幅枚数
            float fWidth = (float)iWidth / (float)myTexture.Width;    // 画像１枚当たりの横幅（float）
            float fHeight = (float)iHeight / (float)myTexture.Height; // 画像１枚当たりの縦幅（float）
            float tx1 = (animeNo % animeWidthMax) * fWidth;
            float ty1 = (animeNo / animeWidthMax) * fHeight;
            float tx2 = tx1 + fWidth;
            float ty2 = ty1 + fHeight;

            // テクスチャ座標を更新する
            TextureCoordinate = new Vector2(tx1, ty1);
        }
    }
}
